/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package pem

import (
	"context"
	"crypto/sha512"
	"io/ioutil"
	"os"
	"path/filepath"
	"testing"
	"time"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"

	"golang.org/x/sys/unix"

	api_v1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes/fake"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
)

func setupSecretLister(client *fake.Clientset) core_v1_listers.SecretLister {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	infFactory := informers.NewSharedInformerFactory(client, 0)
	informer := infFactory.Core().V1().Secrets().Informer()
	lister := infFactory.Core().V1().Secrets().Lister()

	infFactory.Start(ctx.Done())
	cache.WaitForCacheSync(ctx.Done(), informer.HasSynced)
	return lister
}

func TestBasename(t *testing.T) {
	file := File{
		Namespace: "ns",
		Name:      "name",
	}
	if file.Basename() != "ns_name.pem" {
		t.Errorf("File.Basename(): got %s want ns_name.pem",
			file.Basename())
	}

	file = File{
		Namespace: "ns34567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890",
		Name:      "name567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890",
	}
	if len(file.Basename()) > unix.NAME_MAX {
		t.Errorf("len(File.Basename(%s/%s)=%s): got %d want <= %d",
			file.Namespace, file.Name, file.Basename(),
			len(file.Basename()), unix.NAME_MAX)
	}
}

func getTempDir() (string, error) {
	return ioutil.TempDir("", "k8s-crt-dnldr_pem_test")
}

func TestNewFiles(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	if _, err := NewFiles(base, -1, nil); err == nil {
		t.Errorf("NewFiles(getter=nil) error: got nil want non-nil")
	}

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	noSuchDir := filepath.Join(base, "nosuchdir")
	if _, err := NewFiles(noSuchDir, -1, getter); err == nil {
		t.Errorf("NewFiles(no such base) error: got nil want non-nil")
	}
	nonDir := filepath.Join(base, "nondir")
	if err = ioutil.WriteFile(nonDir, []byte{}, os.FileMode(0664)); err != nil {
		t.Fatalf("Cannot write test file %s: %v", nonDir, err)
	}
	if _, err := NewFiles(nonDir, -1, getter); err == nil {
		t.Errorf("NewFiles(non dir base) error: got nil want non-nil")
	}
}

func TestPath(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	files, err := NewFiles(base, -1, getter)
	if err != nil {
		t.Fatalf("NewFiles(): %v", err)
	}
	file := &File{
		Namespace: "ns",
		Name:      "name",
	}
	path := filepath.Join(base, "ns_name.pem")
	if files.Path(file) != path {
		t.Errorf("Files.Path(ns/name): got %s want %s",
			files.Path(file), path)
	}
}

func TestHave(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	files, err := NewFiles(base, -1, getter)
	if err != nil {
		t.Fatalf("NewFiles(): %v", err)
	}
	files.Files["ns/name"] = &File{
		Namespace:       "ns",
		Name:            "name",
		UID:             "1b531285-5d26-4fb9-a1f3-566ac2462e57",
		ResourceVersion: "4711",
	}
	if !files.Have(
		"ns", "name", "1b531285-5d26-4fb9-a1f3-566ac2462e57", "4711") {
		t.Error("Files.Have(): got false want true")
	}
	if !files.Have("ns", "name", "", "4711") {
		t.Error("Files.Have(): got false want true")
	}
	if !files.Have(
		"ns", "name", "1b531285-5d26-4fb9-a1f3-566ac2462e57", "") {
		t.Error("Files.Have(): got false want true")
	}
	if !files.Have("ns", "name", "", "") {
		t.Error("Files.Have(): got false want true")
	}
	if files.Have(
		"foo", "name", "1b531285-5d26-4fb9-a1f3-566ac2462e57", "4711") {
		t.Error("Files.Have(): got true want false")
	}
	if files.Have(
		"ns", "foo", "1b531285-5d26-4fb9-a1f3-566ac2462e57", "4711") {
		t.Error("Files.Have(): got true want false")
	}
}

func TestCheck(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	file := &File{
		Namespace:       "ns",
		Name:            "secret",
		UID:             "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		ResourceVersion: "123456",
	}
	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns",
				Name:            "secret",
				UID:             "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
				ResourceVersion: "123456",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files, err := NewFiles(base, gid, getter)
	if err != nil {
		t.Fatalf("NewFiles(): %v", err)
	}
	files.Files["ns/secret"] = file
	path := files.Path(file)
	bytes := []byte("crtSecret\nkeySecret")
	if err = ioutil.WriteFile(path, bytes, os.FileMode(0640)); err != nil {
		t.Fatalf("ioutil.WriteFile(%s): %+v", path, err)
	}
	if err = os.Chown(path, -1, gid); err != nil {
		t.Fatalf("os.Chown(%s): %+v", path, err)
	}
	info, err := os.Stat(path)
	if err != nil {
		t.Fatalf("os.Stat(%s): %+v", path, err)
	}
	file.Size = info.Size()
	file.ModTime = info.ModTime()
	file.Hash = sha512.Sum512(bytes)

	if !files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got false want true")
	}
	if !files.Check("ns", "secret", "", "123456") {
		t.Error("Check(): got false want true")
	}
	if !files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"") {
		t.Error("Check(): got false want true")
	}
	if !files.Check("ns", "secret", "", "") {
		t.Error("Check(): got false want true")
	}
	if files.Check("foo", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}
	if files.Check("ns", "foo", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}

	// file not found
	cp := filepath.Join(base, "copy")
	if err = os.Rename(path, cp); err != nil {
		t.Fatalf("os.Rename(%s, %s): %+v", path, cp, err)
	}
	if files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}
	if err = os.Rename(cp, path); err != nil {
		t.Fatalf("os.Rename(%s, %s): %+v", cp, path, err)
	}
	if !files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got false want true")
	}

	// change mtime
	now := time.Now()
	if err = os.Chtimes(path, now, now); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}
	if err = os.Chtimes(path, now, file.ModTime); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if !files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got false want true")
	}

	// change size
	bytes = []byte("crtSecret\nkeySecret\n")
	if err = ioutil.WriteFile(path, bytes, os.FileMode(0640)); err != nil {
		t.Fatalf("ioutil.WriteFile(%s): %+v", path, err)
	}
	if err = os.Chtimes(path, now, file.ModTime); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}

	// change hash
	bytes = []byte("keySecret\ncrtSecret")
	if err = ioutil.WriteFile(path, bytes, os.FileMode(0640)); err != nil {
		t.Fatalf("ioutil.WriteFile(%s): %+v", path, err)
	}
	if err = os.Chtimes(path, now, file.ModTime); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}
	bytes = []byte("crtSecret\nkeySecret")
	if err = ioutil.WriteFile(path, bytes, os.FileMode(0640)); err != nil {
		t.Fatalf("ioutil.WriteFile(%s): %+v", path, err)
	}
	if err = os.Chtimes(path, now, file.ModTime); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if !files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got false want true")
	}

	// change mode -- must be g+w
	if err = os.Chmod(path, os.FileMode(0600)); err != nil {
		t.Fatalf("os.Chmod(%s, 0600): %+v", path, err)
	}
	if err = os.Chtimes(path, now, file.ModTime); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got true want false")
	}
	if err = os.Chmod(path, os.FileMode(0640)); err != nil {
		t.Fatalf("os.Chmod(%s, 0600): %+v", path, err)
	}
	if err = os.Chtimes(path, now, file.ModTime); err != nil {
		t.Fatalf("os.Chtimes(%s): %+v", path, err)
	}
	if !files.Check("ns", "secret", "323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
		"123456") {
		t.Error("Check(): got false want true")
	}

	// change gid if possible
	gids, err := os.Getgroups()
	if err != nil && len(gids) > 1 {
		gid2 := -1
		for _, g := range gids {
			if g != gid {
				gid2 = g
				break
			}
		}
		if gid2 != -1 {
			if err = os.Chown(path, -1, gid2); err != nil {
				t.Fatalf("os.Chown(%s): %+v", path, err)
			}
			if files.Check("ns", "secret",
				"323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
				"123456") {
				t.Error("Check(): got true want false")
			}
			if err = os.Chown(path, -1, gid); err != nil {
				t.Fatalf("os.Chown(%s): %+v", path, err)
			}
			if !files.Check("ns", "secret",
				"323ad46e-0fbb-4d0c-a5da-d9daf1ed12df",
				"123456") {
				t.Error("Check(): got false want true")
			}
		}
	}
}

func TestWrite(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "namespace",
				Name:            "name",
				UID:             "73e880f3-d7ca-447f-b95e-88f220d4820e",
				ResourceVersion: "654321",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "missing",
				Name:            "crtField",
				UID:             "29d15ca8-c133-4fc7-ac9d-52441b10f8cf",
				ResourceVersion: "4711",
			},
			Data: map[string][]byte{
				"foo":     []byte("fooVal"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "missing",
				Name:            "keyField",
				UID:             "cdac5d5a-0e88-47dd-b881-99ac4f25222a",
				ResourceVersion: "0815",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"foo":     []byte("fooVal"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "empty",
				Name:            "crtField",
				UID:             "3e3201f2-d09e-49e2-8ee7-7712fb50504f",
				ResourceVersion: "879006",
			},
			Data: map[string][]byte{
				"tls.crt": []byte{},
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "empty",
				Name:            "keyField",
				UID:             "7a6a6a3a-92f6-4fad-9820-8733a0c49760",
				ResourceVersion: "863249",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte{},
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "invalid",
				Name:            "typeField",
				UID:             "6fb9641b-f06d-48eb-98c0-cf1f42870395",
				ResourceVersion: "123456",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("foo/bar"),
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files, err := NewFiles(base, gid, getter)
	if err != nil {
		t.Fatalf("NewFiles(): %v", err)
	}

	found, valid, err := files.Write("namespace", "name",
		"73e880f3-d7ca-447f-b95e-88f220d4820e", "654321")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if !valid {
		t.Error("Write() valid: got false want true")
	}
	if err != nil {
		t.Errorf("Write() err: got %+v want nil", err)
	}
	if !files.Check("namespace", "name",
		"73e880f3-d7ca-447f-b95e-88f220d4820e", "654321") {
		t.Error("Check() after Write(): got false want true")
	}

	// repeat the checks from Check() manually, diagnoses the error
	// if Check() returned false.
	file, ok := files.Files["namespace/name"]
	if !ok {
		t.Fatalf("Write(): map entry not found after call")
	}
	if file.Namespace != "namespace" {
		t.Errorf("file.Namespace after Write(): got %s want "+
			"\"namespace\"", file.Namespace)
	}
	if file.Name != "name" {
		t.Errorf("file.Name after Write(): got %s want \"name\"",
			file.Name)
	}
	if file.UID != "73e880f3-d7ca-447f-b95e-88f220d4820e" {
		t.Errorf("file.UID after Write(): got %s want "+
			"\"73e880f3-d7ca-447f-b95e-88f220d4820e\"", file.UID)
	}
	if file.ResourceVersion != "654321" {
		t.Errorf("file.ResourceVersion after Write(): got %s want "+
			"\"654321\"", file.ResourceVersion)
	}
	bytes := []byte("crtSecret\nkeySecret")
	if hash := sha512.Sum512(bytes); hash != file.Hash {
		t.Error("file.Hash after Write(): mismatch")
	}
	path := files.Path(file)
	info, err := os.Stat(path)
	if err != nil {
		t.Fatalf("os.Stat(%s) after Write(): %+v", path, err)
	}
	if info.Size() != file.Size {
		t.Errorf("os.Stat(%s).Size() after Write(): got %d want %d",
			path, info.Size(), file.Size)
	}
	if info.ModTime() != file.ModTime {
		t.Errorf("os.Stat(%s).ModTime() after Write(): got %s want %s",
			path, info.ModTime(), file.ModTime)
	}
	if info.Mode().Perm()&040 == 0 {
		t.Errorf("os.Stat(%s).Mode() after Write(): g-w", path)
	}
	ustat := &unix.Stat_t{}
	if err = unix.Stat(path, ustat); err != nil {
		t.Fatalf("unix.Stat(%s) after Write(): %+v", path, err)
	} else if int(ustat.Gid) != files.Gid {
		t.Errorf("gid after Write(): got %d want %d", int(ustat.Gid),
			files.Gid)
	}
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s): %+v", path, err)
	} else if sha512.Sum512(contents) != file.Hash {
		t.Error("File contents hash mismatch after Write()")
	}

	found, valid, err = files.Write("namespace", "name", "", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if !valid {
		t.Error("Write() valid: got false want true")
	}
	if err != nil {
		t.Errorf("Write() err: got %+v want nil", err)
	}
	if !files.Check("namespace", "name",
		"73e880f3-d7ca-447f-b95e-88f220d4820e", "654321") {
		t.Error("Check() after Write(): got false want true")
	}

	found, valid, err = files.Write("namespace", "name",
		"73e880f3-d7ca-447f-b95e-88f220d4820e", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if !valid {
		t.Error("Write() valid: got false want true")
	}
	if err != nil {
		t.Errorf("Write() err: got %+v want nil", err)
	}
	if !files.Check("namespace", "name",
		"73e880f3-d7ca-447f-b95e-88f220d4820e", "654321") {
		t.Error("Check() after Write(): got false want true")
	}

	found, valid, err = files.Write("namespace", "name", "", "654321")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if !valid {
		t.Error("Write() valid: got false want true")
	}
	if err != nil {
		t.Errorf("Write() err: got %+v want nil", err)
	}
	if !files.Check("namespace", "name",
		"73e880f3-d7ca-447f-b95e-88f220d4820e", "654321") {
		t.Error("Check() after Write(): got false want true")
	}

	found, valid, err = files.Write("foo", "bar", "", "")
	if found {
		t.Error("Write() found: got true want false")
	}

	found, valid, err = files.Write("missing", "crtField", "", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if valid {
		t.Error("Write() valid: got true want false")
	}
	if err == nil {
		t.Error("Write() error: got nil want non-nil")
	} else if err.Error() != "required field missing" {
		t.Errorf("err msg: want \"required field missing\" got %s",
			err.Error())
	}

	found, valid, err = files.Write("missing", "keyField", "", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if valid {
		t.Error("Write() valid: got true want false")
	}
	if err == nil {
		t.Error("Write() error: got nil want non-nil")
	} else if err.Error() != "required field missing" {
		t.Errorf("err msg: want \"required field missing\" got %s",
			err.Error())
	}

	found, valid, err = files.Write("empty", "crtField", "", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if valid {
		t.Error("Write() valid: got true want false")
	}
	if err == nil {
		t.Error("Write() error: got nil want non-nil")
	} else if err.Error() != "required field empty" {
		t.Errorf("err msg: want \"required field empty\" got %s",
			err.Error())
	}

	found, valid, err = files.Write("empty", "keyField", "", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if valid {
		t.Error("Write() valid: got true want false")
	}
	if err == nil {
		t.Error("Write() error: got nil want non-nil")
	} else if err.Error() != "required field empty" {
		t.Errorf("err msg: want \"required field empty\" got %s",
			err.Error())
	}

	found, valid, err = files.Write("invalid", "typeField", "", "")
	if !found {
		t.Error("Write() found: got false want true")
	}
	if valid {
		t.Error("Write() valid: got true want false")
	}
	if err == nil {
		t.Error("Write() error: got nil want non-nil")
	} else if err.Error() != "invalid type value" {
		t.Errorf("err msg: want \"invalid type value\" got %s",
			err.Error())
	}
}

func TestDelete(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	file := &File{
		Namespace:       "test",
		Name:            "deleteme",
		UID:             "b47b7785-64ff-47fe-8e72-da334856129f",
		ResourceVersion: "4711",
	}
	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns",
				Name:            "secret",
				UID:             "b47b7785-64ff-47fe-8e72-da334856129f",
				ResourceVersion: "4711",
			},
			Data: map[string][]byte{
				"tls.crt": []byte("crtSecret"),
				"tls.key": []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	files, err := NewFiles(base, gid, getter)
	if err != nil {
		t.Fatalf("NewFiles(): %v", err)
	}
	files.Files["test/deleteme"] = file
	path := files.Path(file)
	bytes := []byte("crtSecret\nkeySecret")
	if err = ioutil.WriteFile(path, bytes, os.FileMode(0640)); err != nil {
		t.Fatalf("ioutil.WriteFile(%s): %+v", path, err)
	}
	if err = os.Chown(path, -1, gid); err != nil {
		t.Fatalf("os.Chown(%s): %+v", path, err)
	}
	info, err := os.Stat(path)
	if err != nil {
		t.Fatalf("os.Stat(%s): %+v", path, err)
	}
	file.Size = info.Size()
	file.ModTime = info.ModTime()
	file.Hash = sha512.Sum512(bytes)

	existed, err := files.Delete("test", "deleteme")
	if !existed {
		t.Error("Deleted() existed: got false want true")
	}
	if err != nil {
		t.Errorf("Deleted() error: got %+v want nil", err)
	}
	if files.Check("test", "deleteme", "", "") {
		t.Error("Check() after Delete(): got true want false")
	}
	path = files.Path(file)
	if info, err = os.Stat(path); err == nil {
		t.Errorf("os.Stat(%s) error after Delete(): got nil "+
			"want non-nil", path)
	} else if !os.IsNotExist(err) {
		t.Errorf("os.Stat(%s) error: got %+v want os.IsNotExist()=true",
			path, err)
	}

	existed, err = files.Delete("doesnot", "exist")
	if existed {
		t.Error("Delete() existed: got true want false")
	}
	if err != nil {
		t.Errorf("Delete() error: got %+v want nil", err)
	}

	// file entry is in the map, but the file does not exist
	files.Files["test/deleteme"] = file
	existed, err = files.Delete("test", "deleteme")
	if existed {
		t.Error("Delete() existed: got true want false")
	}
	if err == nil {
		t.Error("Delete() error: got nil want non-nil")
	} else if !os.IsNotExist(err) {
		t.Errorf("Delete() error: got %+v want os.IsNotExist()=true",
			err)
	}
}
